@@ -35,60 +35,6 @@ from selve.util.errors import *
 from selve.util.protocol import ParameterType
 
 
-def _worker(selve: Selve, stop):
-    # Infinite loop to collect all incoming data
-    selve._LOGGER.debug("Reader started")
-    try:
-        while True:
-            if not selve._pauseReader:
-                with selve._readLock:
-                    if not selve._serial.is_open:
-                        selve._serial.open()
-                    if selve._serial.in_waiting > 0:
-                        msg = ""
-                        while True:
-                            response = selve._serial.readline().strip()
-                            msg += response.decode()
-                            if response.decode() == '':
-                                break
-
-                        # do something with the received data
-                        selve.processResponse(msg)
-
-                        # if msg.rstrip() == b' ':
-                        selve._LOGGER.debug(f'Received: {msg}')
-
-            if not selve._pauseWriter:
-                if not selve.txQ.empty():
-                    data: Command = selve.txQ.get_nowait()
-                    commandstr = data.serializeToXML()
-                    selve._LOGGER.debug('Gateway writing: ' + str(commandstr))
-                    try:
-                        with selve._writeLock:
-                            if not selve._serial.is_open:
-                                selve._serial.open()
-                            selve._serial.write(commandstr)
-                            selve._serial.flush()
-                    except Exception as e:
-                        selve._LOGGER.error("error communicating: " + str(e))
-
-                    selve.txQ.task_done()
-
-                    # always sleep after writing
-                    time.sleep(0.5)
-
-            time.sleep(0.01)
-            if stop():
-                selve._LOGGER.debug('Exiting worker loop...')
-                break
-        return True
-    # serial port exceptions, all of these notify that we are in some
-    # serious trouble
-    except serial.SerialException:
-        # log message
-        selve._LOGGER.error('Serial Port RX error')
-
-
 class Selve:
     """Implementation of the serial communication to the Selve Gateway"""
 
@@ -115,7 +61,14 @@ class Selve:
 
         # Flags for enabling reader and writer in the worker thread
         self._pauseReader = False
+        self._pauseReaderEvent = threading.Event()
         self._pauseWriter = False
+        self._pauseWriterEvent = threading.Event()
+        self._pauseWorker = False
+        self._pauseWorkerEvent = threading.Event()
+        
+        # The worker thread
+        self.workerTask = None
 
         # Port where the Selve gateway was found
         self._port = port
@@ -127,15 +80,77 @@ class Selve:
         self._LOGGER = logger
 
         self._setup()
-        
-        self.readLoopTask = threading.Thread(target=_worker, args=(self, lambda: self._stopThread))
-        self.readLoopTask.daemon = False
-        self.readLoopTask.start()
+        self.startWorker()
 
         if discover:
             self._LOGGER.info("Discovering devices")
             self.discover()
 
+    
+    def _worker(self, selve: Selve, _writeLock, _readLock):
+        # Infinite loop to collect all incoming data
+        self._LOGGER.debug("Worker started")
+        try:
+            while True:
+                if not selve._pauseWorker:
+                    selve._pauseWorkerEvent.clear()
+                    if not selve._pauseReader:
+                        selve._pauseReaderEvent.clear()
+                        with _readLock():
+                            if not self._serial.is_open:
+                                self._serial.open()
+                            if selve._serial.in_waiting > 0:
+                                msg = ""
+                                while True:
+                                    response = selve._serial.readline().strip()
+                                    msg += response.decode()
+                                    if response.decode() == '':
+                                        break
+
+                                # do something with the received data
+                                selve.processResponse(msg)
+
+                                # if msg.rstrip() == b' ':
+                                selve._LOGGER.debug(f'Received: {msg}')
+                    else:
+                        selve._pauseReaderEvent.set()
+                        selve._LOGGER.debug("Reader stopped")
+                    if not selve._pauseWriter:
+                        selve._pauseWriterEvent.clear()
+                        if not selve.txQ.empty():
+                            data: Command = selve.txQ.get_nowait()
+                            commandstr = data.serializeToXML()
+                            selve._LOGGER.debug('Gateway writing: ' + str(commandstr))
+                            try:
+                                with _writeLock():
+                                    if not self._serial.is_open:
+                                        self._serial.open()
+                                    selve._serial.write(commandstr)
+                                    selve._serial.flush()
+                            except Exception as e:
+                                selve._LOGGER.error("error communicating: " + str(e))
+
+                            selve.txQ.task_done()
+
+                            # always sleep after writing
+                            time.sleep(1)
+                    else:
+                        selve._pauseWriterEvent.set()
+                        selve._LOGGER.debug("Writer stopped")
+                else:
+                    selve._pauseWorkerEvent.set()
+                time.sleep(0.01)
+                if selve._stopThread:
+                    selve._LOGGER.debug('Exiting worker loop...')
+                    break
+            return True
+        # serial port exceptions, all of these notify that we are in some
+        # serious trouble
+        except serial.SerialException:
+            # log message
+            self._LOGGER.error('Serial Port RX error')
+
+
     def _setup(self):
         self._LOGGER.info("Setup")
         self.rxQ = queue.Queue()
@@ -196,11 +211,32 @@ class Selve:
             self._LOGGER.error("No gateway on comports found!")
             raise PortError
 
+    def startWorker(self):
+        self._LOGGER.debug("Starting worker")
+        self._pauseReader = False
+        self._pauseWriter = False
+        self._pauseWorker = False
+        self._stopThread = False
+        self.workerTask = threading.Thread(target=self._worker, args=(self, lambda: self._writeLock, lambda: self._readLock), daemon=True)
+        self.workerTask.start()
+
+    def stopWorker(self):
+        self._LOGGER.debug("Stopping worker")
+        self._pauseReader = True
+        self._pauseWriter = True
+        self._pauseWorker = True
+        self._stopThread = True
+        if self.workerTask is not None and self.workerTask.is_alive():
+            self.workerTask.join()
+        self.workerTask = None
+
+
     def stopGateway(self):
         # wait for the rx/tx thread to end, these need to be gathered to
         # collect all the exceptions
+        self._LOGGER.debug("Preparing for termination")
         self._stopThread = True
-        self.readLoopTask.join()
+        self.workerTask.join()
         # close the serial port, do the cleanup
         if self._serial.is_open:
             self._serial.close()
@@ -224,12 +260,13 @@ class Selve:
                     self._serial.open()
                 self._serial.write(commandstr)
                 self._serial.flush()
-                time.sleep(0.5)
+                # always sleep after writing
+                time.sleep(1)
         except Exception as e:
             self._LOGGER.error("error communicating: " + str(e))
 
     def processResponse(self, xmlstr):
-        """Processes an XML String into a response object."""
+        """Processes an XML String into a response object. Returns False if something went wrong or the gateway returned an error."""
         # check which command was received
         # do something with the data
         # return the ready to eat response
@@ -263,7 +300,7 @@ class Selve:
                     or isinstance(response, SenderEventResponse) \
                     or isinstance(response, LogEventResponse) \
                     or isinstance(response, DutyCycleResponse):
-                return self.processEventResponse(response)
+                self.processEventResponse(response)
             if isinstance(response, CommandResultResponse)\
                     or isinstance(response, IveoResultResponse):
                 #update device values
@@ -281,7 +318,7 @@ class Selve:
         if hasattr(obj, "methodResponse"):
             return ErrorResponse(obj.methodResponse.fault.array.string.cdata, obj.methodResponse.fault.array.int.cdata)
         else:
-            raise CommunicationError()
+            return False
 
     def create_response(self, obj):
         if hasattr(obj, "methodResponse"):
@@ -497,12 +534,7 @@ class Selve:
         return MethodResponse(methodName, flat_params_list)
 
     def executeCommand(self, command: Command):
-        if not self.readLoopTask.is_alive():
-            self._stopThread = False
-            self.readLoopTask = threading.Thread(target=_worker, args=(self, lambda: self._stopThread))
-            self.readLoopTask.daemon = False
-            self.readLoopTask.start()
-
+        self._pauseWorker = False
         self.txQ.put_nowait(command)
 
 
@@ -516,12 +548,15 @@ class Selve:
 
     def _executeCommandSyncWithResponse(self, command: Command):
 
-        self._pauseWriter = True
-        self._pauseReader = True
-        self._pauseWorker = True
+        self.stopWorker()
+
         if not self._serial.is_open:
             self._serial.open()
 
+        #guarantee exclusivity
+        #self._serial.reset_input_buffer()
+        #self._serial.reset_output_buffer()
+
         self._sendCommandToGateway(command)
 
         start_time = time.time()
@@ -536,11 +571,8 @@ class Selve:
                             break
                     # if msg.rstrip() == b' ':
                     self._LOGGER.debug(f'Received: {msg}')
-
-                    self._pauseReader = False
-                    self._pauseWriter = False
-                    self._pauseWorker = False
-
+                    self.startWorker()
+    
                     resp = self.processResponse(msg)
                     
                     if (resp == False):
@@ -550,21 +582,21 @@ class Selve:
                     if isinstance(resp, ErrorResponse):
                         self._LOGGER.error(resp.message)
                         # retry
-                        try:
-                            return self._executeCommandSyncWithResponse(command)
-                        except GatewayError as e:
-                            raise e
-                    else:
-                        #time.sleep(0.5)
+                        return False
+
                         return resp
-                if time.time() - start_time > 10:
-                    time.sleep(0.05)
-                    return None
+                    # When no data is waiting in the input buffer after 10s we can assume, the message was not correctly sent or no input is necessary
+                    if time.time() - start_time < 10:
+                        time.sleep(0.05)
+                    else:
+                        self.startWorker()
+                        return False
+        
 
 
     def discover(self):
 
-        self._stopThread = True
+        self._pauseWorker = True
 
         if self.gatewayReady():
             iveoIds: IveoGetIdsResponse = self.executeCommandSyncWithResponse(IveoGetIds())
@@ -663,10 +695,8 @@ class Selve:
                 device.sun2Analog = config.sun2Analog
                 device.sun3Analog = config.sun3Analog
                 self.addOrUpdateDevice(device, SelveTypes.SENSIM)
-        self._stopThread = False
-        self.readLoopTask = threading.Thread(target=_worker, args=(self, lambda: self._stopThread))
-        self.readLoopTask.daemon = False
-        self.readLoopTask.start()
+        
+        self._pauseWorker = False
         self.list_devices()
 
 
